#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import os.path
import datetime
import sys
import logging
import numpy
import argparse
import pandas

# Import required src

from usienarl import Experiment


def softmax(sequence):
    """
    Compute the softmax of a sequence of values, i.e. the resulting probability distribution over that sequence.

    :param sequence: the sequence of which to compute the softmax
    :return: the probability distribution over the sequence
    """
    # Make sure the sequence is a numpy array
    if not isinstance(sequence, numpy.ndarray):
        sequence = numpy.array(sequence)
    # Make sure the length of the shape of the given array is 2
    assert len(sequence.shape) == 2
    # Get the element with max value in the given array as the normalization factor
    normalization_factor = numpy.max(sequence, axis=1)
    # Increase the shape size of the normalization factor to allow broadcasting
    normalization_factor = normalization_factor[:, numpy.newaxis]
    # Apply the normalization
    numerator = numpy.exp(sequence - normalization_factor)
    # Compute the denominator by summing all the normalized numerator elements
    denominator = numpy.sum(numerator, axis=1)
    # Increase the shape size of the denominator to allow broadcasting
    denominator = denominator[:, numpy.newaxis]
    # Return the softmax result probability distribution
    return numerator / denominator


def _positive_int(value: str) -> int:
    """
    Positive integer check/conversion rule for arg-parse.

    :param value: the value to check if it's integer
    :return: the int value if positive
    """
    int_value: int = int(value)
    if int_value <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return int_value


def command_line_parse():
    """
    Parse command line using arg-parse and get built-in run experiment user data.

    :return: the parsed arguments: workspace directory, experiment iterations, CUDA devices, optional render during training flag, optional render during validation flag, optional render during testing flag
    """
    # Parse depending on the boolean watch flag
    parser = argparse.ArgumentParser()
    parser.add_argument("workspace_directory", type=str)
    parser.add_argument("experiment_iterations", type=_positive_int)
    parser.add_argument("CUDA_devices", type=str)
    parser.add_argument("-render_during_training", action="store_true")
    parser.add_argument("-render_during_validation", action="store_true")
    parser.add_argument("-render_during_testing", action="store_true")
    args: dict = vars(parser.parse_args())
    return args["workspace_directory"], args["experiment_iterations"], args["CUDA_devices"], args["render_during_training"], args["render_during_validation"], args["render_during_testing"]


def run_experiment(logger: logging.Logger,
                   experiment: Experiment,
                   file_name: str, workspace_path: str,
                   training_volleys_episodes: int,
                   validation_volleys_episodes: int,
                   training_validation_volleys: int,
                   test_volleys_episodes: int,
                   test_volleys: int,
                   episode_length: int,
                   parallel: int = 1,
                   plots_dpi: int = 150,
                   render_during_training: bool = False, render_during_validation: bool = False, render_during_test: bool = False,
                   saves_to_keep: int = 1,
                   restore_path: str = None,
                   iterations: int = 1):
    """
    Run the given experiment with the given parameters. It automatically creates all the directories to store the
    experiment results, summaries and meta-graphs for each iteration.

    :param logger: the logger used to record information, warnings and errors
    :param experiment: the experiment to run
    :param file_name: the name of the experiment script
    :param workspace_path: the path of workspace directory
    :param training_volleys_episodes: the number of episodes per training volley
    :param validation_volleys_episodes: the number of episodes per validation volley
    :param training_validation_volleys: the number of training and validation volleys
    :param test_volleys_episodes: the number of episodes per test volley
    :param test_volleys: the number of test volleys
    :param episode_length: the maximum number of steps allowed for each episode in any volley
    :param parallel: the number of parallel episodes to run in the experiment
    :param plots_dpi: the dpi (quality) of plots saved during experiment run
    :param render_during_training: flag to render or not the environment during training
    :param render_during_validation: flag to render or not the environment during validation
    :param render_during_test: flag to render or not the environment during test
    :param saves_to_keep: the number of checkpoint metagraphs to keep
    :param restore_path: the optional checkpoint path from which to restore the model
    :param iterations: the number of iterations of the exact same experiment to run
    :return: the list of saved metagraph paths of each iteration of the experiment
    """
    # Generate the workspace directory if not defined
    workspace_directory = os.path.dirname(workspace_path)
    if not os.path.isdir(workspace_directory):
        try:
            os.makedirs(workspace_directory)
        except FileExistsError:
            pass
    # Generate the file directory with date if not defined
    file_path: str = workspace_path + "/" + os.path.splitext(os.path.basename(file_name))[0] + "_" + datetime.date.today().strftime("%Y-%m-%d")
    if not os.path.isdir(file_path):
        try:
            os.makedirs(file_path)
        except FileExistsError:
            pass
    # Generate a directory for the current experiment if not defined
    experiment_path: str = file_path + "/" + experiment.name
    if not os.path.isdir(experiment_path):
        try:
            os.makedirs(experiment_path)
        except FileExistsError:
            pass
    # Generate a directory for each iteration of the experiment if not defined
    # Prepare a table in which to store all results and a list of metagraph paths
    experiment_results_table: dict = dict()
    for iteration in range(iterations):
        iteration_path: str = experiment_path
        # Set the iteration number to -1 if there is only one iteration
        if iterations <= 1:
            iteration = -1
        if iteration > -1:
            iteration_path = experiment_path + "/" + str(iteration)
            if not os.path.isdir(iteration_path):
                try:
                    os.makedirs(iteration_path)
                except FileExistsError:
                    pass
        # Generate a metagraph, summary and plots directory for each iteration of experiment if not defined
        plots_path: str = iteration_path + "/plots"
        if not os.path.isdir(plots_path):
            try:
                os.makedirs(plots_path)
            except FileExistsError:
                pass
        save_path: str = iteration_path + "/metagraph"
        if not os.path.isdir(save_path):
            try:
                os.makedirs(save_path)
            except FileExistsError:
                pass
        summary_path: str = iteration_path + "/summary"
        if not os.path.isdir(summary_path):
            try:
                os.makedirs(summary_path)
            except FileExistsError:
                pass
        # Setup of the logger for the current experiment
        # Reset logger handlers for current experiment
        logger.handlers = []
        # Generate a console and a file handler for the logger
        console_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
        file_handler: logging.FileHandler = logging.FileHandler(iteration_path + "/info.log", "w+")
        # Set handlers properties
        console_handler.setLevel(logging.DEBUG)
        file_handler.setLevel(logging.DEBUG)
        formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        console_handler.setFormatter(formatter)
        file_handler.setFormatter(formatter)
        # Add the handlers to the logger
        logger.addHandler(console_handler)
        logger.addHandler(file_handler)
        # Setup the experiment
        if experiment.setup(logger, plots_path, plots_dpi, parallel, summary_path, save_path, saves_to_keep, iteration):
            # Train the agent
            experiment.train(logger, training_volleys_episodes, validation_volleys_episodes, training_validation_volleys, episode_length,
                             render_during_training, render_during_validation, restore_path)
            # Test the agent
            experiment.test(logger, test_volleys_episodes, test_volleys, episode_length, save_path, render_during_test)
            # Save the result of the iteration in the table
            experiment_results_table[iteration] = ("YES" if experiment.successful else "NO",
                                                   numpy.round(experiment.avg_test_avg_total_reward, 3),
                                                   numpy.round(experiment.max_test_avg_total_reward, 3),
                                                   numpy.round(experiment.avg_test_avg_scaled_reward, 3),
                                                   numpy.round(experiment.max_test_avg_scaled_reward, 3),
                                                   experiment.trained_episodes)
    # Print the results in both a results log file and in the console
    # Reset logger handlers for current experiment
    logger.handlers = []
    results_console_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
    results_file_handler: logging.FileHandler = logging.FileHandler(experiment_path + "/results.log", "w+")
    results_console_handler.setLevel(logging.INFO)
    results_file_handler.setLevel(logging.INFO)
    results_console_formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    results_file_formatter: logging.Formatter = logging.Formatter("%(message)s")
    results_console_handler.setFormatter(results_console_formatter)
    results_file_handler.setFormatter(results_file_formatter)
    # Add the file handler to the logger and print table of results
    logger.addHandler(results_file_handler)
    # Write results
    logger.info(experiment.name + " results:\n")
    with pandas.option_context('display.max_rows', 500, 'display.max_columns', 500, 'display.width', 500):
        data_frame = pandas.DataFrame.from_dict(experiment_results_table, orient="index")
        data_frame.index.name = 'Experiment iteration'
        data_frame.columns = ['Success', 'Average total reward', 'Max total reward', 'Average scaled reward', 'Max scaled reward', 'Trained episodes']
        # Print data frame removing index if there is only one iteration
        if iterations <= 1:
            logger.info(data_frame.to_string(index=False))
        else:
            logger.info(data_frame.to_string())
    logger.info("\n\n")
    # Also write results to csv file
    data_frame.to_csv(experiment_path + "/results_table.csv")
    # Add also the console handler and print verbose results
    logger.addHandler(results_console_handler)
    # Group all the final scores, best scores and training episode counts
    average_total_reward_list: [] = []
    average_scaled_reward_list: [] = []
    max_total_reward_list: [] = []
    max_scaled_reward_list: [] = []
    trained_episodes_list: [] = []
    success_list: [] = []
    for experiment_iteration, result in experiment_results_table.items():
        success_list.append(True if result[0] == "YES" else False)
        average_total_reward_list.append(result[1])
        max_total_reward_list.append(result[2])
        average_scaled_reward_list.append(result[3])
        max_scaled_reward_list.append(result[4])
        trained_episodes_list.append(result[5])
    max_average_total_reward_index: int = int(numpy.argmax(average_total_reward_list))
    max_max_total_reward_index: int = int(numpy.argmax(max_total_reward_list))
    max_average_scaled_reward_index: int = int(numpy.argmax(average_scaled_reward_list))
    max_max_scaled_reward_index: int = int(numpy.argmax(max_scaled_reward_list))
    min_training_episodes_index: int = int(numpy.argmin(trained_episodes_list))
    # Print best results for each indicator
    logger.info("Maximum average total reward over all " + str(test_volleys * test_volleys_episodes) + " test episodes: " + str("%.3f" % numpy.max(average_total_reward_list)) + " at experiment number " + str(max_average_total_reward_index))
    logger.info("Maximum max total score over " + str(test_volleys_episodes) + " test episodes: " + str("%.3f" % numpy.max(max_total_reward_list)) + " at experiment number " + str(max_max_total_reward_index))
    logger.info("Maximum average scaled reward over all " + str(test_volleys * test_volleys_episodes) + " test episodes: " + str("%.3f" % numpy.max(average_scaled_reward_list)) + " at experiment number " + str(max_average_scaled_reward_index))
    logger.info("Maximum max scaled score over " + str(test_volleys_episodes) + " test episodes: " + str("%.3f" % numpy.max(max_scaled_reward_list)) + " at experiment number " + str(max_max_scaled_reward_index))
    logger.info("Minimum training episodes for validation: " + str(numpy.min(trained_episodes_list)) + " at experiment number " + str(min_training_episodes_index))
    # Print average results for each indicator
    average_average_total_reward: float = numpy.average(average_total_reward_list)
    logger.info("Average of average total reward over all experiments " + str("%.3f" % average_average_total_reward))
    average_max_total_reward: float = numpy.average(max_total_reward_list)
    logger.info("Average of max total reward over all experiments: " + str("%.3f" % average_max_total_reward))
    average_average_scaled_reward: float = numpy.average(average_scaled_reward_list)
    logger.info("Average of average scaled reward over all experiments " + str("%.3f" % average_average_scaled_reward))
    average_max_scaled_reward: float = numpy.average(max_scaled_reward_list)
    logger.info("Average of max scaled reward over all experiments: " + str("%.3f" % average_max_scaled_reward))
    average_training_episodes: float = numpy.average(trained_episodes_list)
    logger.info("Average of training episodes over all experiments: " + str("%.3f" % average_training_episodes))
    success_percentage: float = numpy.round(100.0 * numpy.count_nonzero(success_list) / len(success_list), 2)
    logger.info("Success percentage over all experiments: " + str("%.2f" % success_percentage))
    # Print standard deviation for each indicator
    # Compute a list of averages for array difference
    average_total_reward_average_list = numpy.full(len(average_total_reward_list), average_average_total_reward)
    max_total_reward_average_list = numpy.full(len(max_total_reward_list), average_max_total_reward)
    average_scaled_reward_average_list = numpy.full(len(average_scaled_reward_list), average_average_scaled_reward)
    max_scaled_reward_average_list = numpy.full(len(max_scaled_reward_list), average_max_scaled_reward)
    training_episodes_average_list = numpy.full(len(trained_episodes_list), average_training_episodes)
    # Compute the actual standard deviation for each indicator
    stdv_average_total_reward: float = numpy.sqrt(numpy.sum(numpy.power(average_total_reward_list - average_total_reward_average_list, 2)) / len(average_total_reward_list))
    logger.info("Standard deviation of average total reward over all experiments: " + str("%.3f" % stdv_average_total_reward) + ", which is a " + str("%.2f" % ((100 * stdv_average_total_reward) / average_average_total_reward)) + "% of the average")
    stdv_max_total_reward: float = numpy.sqrt(numpy.sum(numpy.power(max_total_reward_list - max_total_reward_average_list, 2)) / len(max_total_reward_list))
    logger.info("Standard deviation of max total reward over all experiments " + str("%.3f" % stdv_max_total_reward) + ", which is a " + str("%.2f" % ((100 * stdv_max_total_reward) / average_max_total_reward)) + "% of the average")
    stdv_average_scaled_reward: float = numpy.sqrt(numpy.sum(numpy.power(average_scaled_reward_list - average_scaled_reward_average_list, 2)) / len(average_scaled_reward_list))
    logger.info("Standard deviation of average scaled reward over all experiments: " + str("%.3f" % stdv_average_scaled_reward) + ", which is a " + str("%.2f" % ((100 * stdv_average_scaled_reward) / average_average_scaled_reward)) + "% of the average")
    stdv_max_scaled_reward: float = numpy.sqrt(numpy.sum(numpy.power(max_scaled_reward_list - max_scaled_reward_average_list, 2)) / len(max_scaled_reward_list))
    logger.info("Standard deviation of max scaled reward over all experiments " + str("%.3f" % stdv_max_scaled_reward) + ", which is a " + str("%.2f" % ((100 * stdv_max_scaled_reward) / average_max_scaled_reward)) + "% of the average")
    stdv_training_episodes: float = numpy.sqrt(numpy.sum(numpy.power(trained_episodes_list - training_episodes_average_list, 2)) / len(trained_episodes_list))
    logger.info("Standard deviation of training episodes over all experiments: " + str("%.3f" % stdv_training_episodes) + ", which is a " + str("%.2f" % ((100 * stdv_training_episodes) / average_training_episodes)) + "% of the average")



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import random
import tensorflow

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.models import DuelingDeepQLearning


class DuelingDeepQLearningAgentEpsilonGreedy(Agent):
    """
    Dueling Double Deep Q-Learning agent with epsilon greedy exploration policy.
    It is supplied with a Dueling Double Deep Q-Learning model (DDDQN).

    The weight copy step interval defines after how many steps per interval the target network weights should be
    updated with the main network weights.

    The batch size define how many steps from the prioritized experience replay built-in buffer should be fed into
    the model when updating. The agent uses a 1-step temporal difference algorithm, i.e. it is updated at every step.

    During training, agent will chose a random action with probability epsilon. Such probability is called exploration
    rate and decays at every completed episode of its decay value, while remaining bounded by its defined max and min
    values.

    Note: default model is suited for simple environment without complex dynamics (for example, without an action mask).
    It is advised to implement a custom model for research task using default ones as templates.
    """

    def __init__(self,
                 name: str,
                 model: DuelingDeepQLearning,
                 summary_save_every_steps: int = 500,
                 weight_copy_every_steps: int = 100,
                 batch_size: int = 64,
                 exploration_rate_max: float = 1.0, exploration_rate_min: float = 0.001,
                 exploration_rate_decay: float = 0.001):
        # Define internal attributes
        self._model: DuelingDeepQLearning = model
        self._exploration_rate_max: float = exploration_rate_max
        self._exploration_rate_min: float = exploration_rate_min
        self._exploration_rate_decay: float = exploration_rate_decay
        self._summary_save_every_steps: int = summary_save_every_steps
        self._weight_copy_every_steps: int = weight_copy_every_steps
        self._batch_size: int = batch_size
        # Define empty attributes
        self._last_training_step: int = 0
        self._last_summary_save_step: int = 0
        self._last_weight_copy_step: int = 0
        self._exploration_rate: float or None = None
        # Generate base agent
        super(DuelingDeepQLearningAgentEpsilonGreedy, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        # Generate the model and return a flag stating if generation was successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    self._parallel,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset internal attributes
        self._last_training_step = 0
        self._last_summary_save_step = 0
        self._last_weight_copy_step = 0
        # Initialize the model
        self._model.initialize(logger, session)
        # Reset exploration rate to its starting value (the max)
        self._exploration_rate = self._exploration_rate_max
        # Run the weight copy operation to uniform main and target networks
        self._model.copy_weights(session)

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   warmup_step: int, warmup_episode: int):
        # Act randomly
        return interface.sample_agent_action(logger, session)

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  train_step: int, train_episode: int):
        # Choose an action according to the epsilon greedy approach: best action predicted by the model or random action
        action: numpy.ndarray = self._model.get_action_with_highest_q_value(session, agent_observation_current, interface.possible_agent_actions(logger, session))
        if self._exploration_rate > 0:
            random_action: numpy.ndarray = interface.sample_agent_action(logger, session)
            for i in range(self._parallel):
                if random.uniform(0, 1.0) < self._exploration_rate:
                    action[i] = random_action[i]
        # Return the exploration action
        return action

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int):
        # Act with the best policy according to the model
        return self._model.get_action_with_highest_q_value(session, agent_observation_current, interface.possible_agent_actions(logger, session))

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        # Adjust the next observation if the episode is done
        if self._observation_space_type == SpaceType.discrete:
            agent_observation_next[episode_done] = 0
        else:
            agent_observation_next[episode_done] = numpy.zeros(self._observation_space_shape, dtype=float)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), agent_observation_next.copy(), episode_done.copy())

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        # Adjust the next observation if the episode is done
        if self._observation_space_type == SpaceType.discrete:
            agent_observation_next[episode_done] = 0
        else:
            agent_observation_next[episode_done] = numpy.zeros(self._observation_space_shape, dtype=float)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), agent_observation_next.copy(), episode_done.copy())
        # Get the number of steps actually done in the environment according to parallelization
        for step in range(self._last_training_step, train_step):
            # After each weight step interval update the target network weights with the main network weights
            if step - self._last_weight_copy_step >= self._weight_copy_every_steps:
                self._model.copy_weights(session)
                # Update last weight copy step
                self._last_weight_copy_step = step
            # Update the model and get current loss and absolute errors
            loss, absolute_errors = self._model.update(session, self._model.buffer.get(self._batch_size))
            # Update the buffer with absolute error
            self._model.buffer.update(absolute_errors)
            # Save the summary at the current step if it is the appropriate time and if required
            if self._summary_writer is not None:
                if step - self._last_summary_save_step >= self._summary_save_every_steps:
                    # Generate the tensorflow summary on the loss and add it to the writer
                    summary = tensorflow.Summary()
                    summary.value.add(tag="loss", simple_value=loss)
                    self._summary_writer.add_summary(summary, step)
                    # Update last summary save step
                    self._last_summary_save_step = step
        # Update last training step
        self._last_training_step = train_step

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()
        # Decrease the exploration rate by its decay value
        self._exploration_rate = max(self._exploration_rate_min, self._exploration_rate - self._exploration_rate_decay)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_step: int, inference_episode: int):
        pass

    @property
    def saved_variables(self):
        # Return the trainable variables of the agent model in experiment/agent scope
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        # Return the amount of warmup episodes required by the model
        return self._model.warmup_steps



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import scripts

from .tabular_sarsa_agent_epsilon_greedy import TabularSARSAAgentEpsilonGreedy
from .tabular_sarsa_agent_boltzmann import TabularSARSAAgentBoltzmann
from .tabular_sarsa_agent_dirichlet import TabularSARSAAgentDirichlet
from .tabular_expected_sarsa_agent_epsilon_greedy import TabularExpectedSARSAAgentEpsilonGreedy
from .tabular_expected_sarsa_agent_boltzmann import TabularExpectedSARSAAgentBoltzmann
from .tabular_expected_sarsa_agent_dirichlet import TabularExpectedSARSAAgentDirichlet
from .tabular_q_learning_agent_epsilon_greedy import TabularQLearningAgentEpsilonGreedy
from .tabular_q_learning_agent_boltzmann import TabularQLearningAgentBoltzmann
from .tabular_q_learning_agent_dirichlet import TabularQLearningAgentDirichlet

from .ddpg_agent import DDPGAgent

from .deep_sarsa_agent_epsilon_greedy import DeepSARSAAgentEpsilonGreedy
from .deep_sarsa_agent_boltzmann import DeepSARSAAgentBoltzmann
from .deep_sarsa_agent_dirichlet import DeepSARSAAgentDirichlet
from .deep_expected_sarsa_agent_epsilon_greedy import DeepExpectedSARSAAgentEpsilonGreedy
from .deep_expected_sarsa_agent_boltzmann import DeepExpectedSARSAAgentBoltzmann
from .deep_expected_sarsa_agent_dirichlet import DeepExpectedSARSAAgentDirichlet
from .deep_q_learning_agent_epsilon_greedy import DeepQLearningAgentEpsilonGreedy
from .deep_q_learning_agent_boltzmann import DeepQLearningAgentBoltzmann
from .deep_q_learning_agent_dirichlet import DeepQLearningAgentDirichlet
from .double_deep_q_learning_agent_epsilon_greedy import DoubleDeepQLearningAgentEpsilonGreedy
from .double_deep_q_learning_agent_boltzmann import DoubleDeepQLearningAgentBoltzmann
from .double_deep_q_learning_agent_dirichlet import DoubleDeepQLearningAgentDirichlet
from .dueling_deep_q_learning_agent_epsilon_greedy import DuelingDeepQLearningAgentEpsilonGreedy
from .dueling_deep_q_learning_agent_boltzmann import DuelingDeepQLearningAgentBoltzmann
from .dueling_deep_q_learning_agent_dirichlet import DuelingDeepQLearningAgentDirichlet

from .vpg_agent import VPGAgent
from .ppo_agent import PPOAgent

from .random_agent import RandomAgent



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy

# Import required src

from usienarl import Environment, SpaceType


class Interface:
    """
    Base interface abstract class to define interaction between agent and environment.

    An interface allows to translate the environment state and actions to the agent observations and actions.
    The default interface is the pass-through interface, implementing a simple fully observable interface for the
    environment. The environment is specific for each interface.

    You should always define your own interface or use a pass-trough, the base class cannot be used as-is.
    """

    def __init__(self,
                 environment: Environment):
        # Make sure parameters are valid
        assert(environment is not None)
        # Define internal attributes
        self._environment: Environment = environment

    def sample_agent_action(self,
                            logger: logging.Logger,
                            session) -> numpy.ndarray:
        """
        Sample a random action as seen by the agent, i.e. a random action in the environment translated in agent action.
        The way the action is sampled depends on the environment.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: the random action as seen by the agent, wrapped in a numpy array
        """
        # Sample an action from the environment
        environment_action: numpy.ndarray = self._environment.sample_action(logger, session)
        # Translate it to agent action and return it
        return self.environment_action_to_agent_action(logger, session, environment_action)

    def agent_action_to_environment_action(self,
                                           logger: logging.Logger,
                                           session,
                                           agent_action: numpy.ndarray) -> numpy.ndarray:
        """
        Translate the given agent action to the respective environment action, both wrapped in a numpy array to allow
        parallelization.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running
        :param agent_action: the action as seen by the agent, wrapped in a numpy array
        :return: the action as seen by the environment relative to the given agent action, wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def environment_action_to_agent_action(self,
                                           logger: logging.Logger,
                                           session,
                                           environment_action: numpy.ndarray) -> numpy.ndarray:
        """
        Translate the given environment action to the respective agent action, both wrapped in a numpy array to allow
        parallelization.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running
        :param environment_action: the action as seen by the environment, wrapped in a numpy array
        :return: the action as seen by the agent relative to the given environment action, wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def environment_state_to_observation(self,
                                         logger: logging.Logger,
                                         session,
                                         environment_state: numpy.ndarray) -> numpy.ndarray:
        """
        Translate the given environment state to the respective agent state, i.e. observation, both wrapped in a
        numpy array to allow parallelization.

        :param logger: the logger used to print the interface information, warnings and errors
        :param session: the session of tensorflow currently running
        :param environment_state: the state as seen by the environment, wrapped in a numpy array
        :return: the state as seen (observed) by the agent relative to the given environment state, wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def possible_agent_actions(self,
                               logger: logging.Logger,
                               session) -> []:
        """
        Get a list of all actions' indexes possible at the current states of the environment if the action space
        is discrete, translated into agent actions.
        Get a list containing the lower and the upper bounds at the current states of the environment, each
        wrapped in numpy array with the shape of the  action space, translated into agent actions.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: a list of indices containing the possible actions or a list of upper and lower bounds arrays, translated into agent actions
        """
        # Translate each action in each batch of actions from environment to agent
        batch_environment_possible_actions: [] = self._environment.possible_actions(logger, session)
        batch_agent_possible_actions: [] = []
        for environment_possible_actions in batch_environment_possible_actions:
            agent_possible_actions: [] = []
            for environment_action in environment_possible_actions:
                # If single environment action is discrete append it as a number, otherwise as an array (the lower/upper boundary)
                if self.agent_action_space_type == SpaceType.discrete:
                    agent_possible_actions.append(self.environment_action_to_agent_action(logger, session, numpy.array(environment_action)).tolist())
                else:
                    agent_possible_actions.append(self.environment_action_to_agent_action(logger, session, numpy.array(environment_action)))
            batch_agent_possible_actions.append(agent_possible_actions)
        # Return the agent version of the possible actions
        return batch_agent_possible_actions

    @property
    def environment(self) -> Environment:
        """
        The environment associated with the interface.
        """
        return self._environment

    @property
    def observation_space_type(self) -> SpaceType:
        """
        The type of the observation space of the agent.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def observation_space_shape(self) -> ():
        """
        The shape of the observation space of the agent.
        Note: it may differ from the environment's state space shape.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def agent_action_space_type(self) -> SpaceType:
        """
        The type of the action space of the agent.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def agent_action_space_shape(self) -> ():
        """
        The shape of the action space of the agent.
        Note: it may differ from the environment's action space shape.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()



# Import packages

import tensorflow
import numpy
import math

# Import required src

from usienarl import SpaceType, Config, Model


class Buffer:
    """
    A buffer for storing trajectories experienced by a DDPG agent interacting with the environment, it is a simple FIFO
    experience replay buffer.
    """
    def __init__(self,
                 capacity: int,
                 parallel_amount: int,
                 observation_space_shape: (),
                 action_space_shape: ()):
        # Define buffer attributes
        self._capacity: int = capacity
        self._parallel_amount: int = parallel_amount
        # Define buffer empty attributes
        self._pointer: int = 0
        self._size: int = 0
        self._observations_current: numpy.ndarray = numpy.zeros((self._capacity, *observation_space_shape), dtype=float)
        self._observations_next: numpy.ndarray = numpy.zeros((self._capacity, *observation_space_shape), dtype=float)
        self._actions: numpy.ndarray = numpy.zeros((self._capacity, *action_space_shape), dtype=float)
        self._rewards: numpy.ndarray = numpy.zeros(self._capacity, dtype=float)
        self._episode_done_flags: numpy.ndarray = numpy.zeros(self._capacity, dtype=float)
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    def store(self,
              observation_current: numpy.ndarray,
              action: numpy.ndarray,
              reward: numpy.ndarray,
              observation_next: numpy.ndarray,
              episode_done: numpy.ndarray):
        """
        Store the time-step in the buffer.

        :param observation_current: the current observation to store in the buffer
        :param action: the last action to store in the buffer
        :param reward: the reward obtained from the action at the current state to store in the buffer
        :param observation_next: the next observation to store in the buffer
        :param episode_done: whether or not this time-step was the last of the episode
        """
        # Serialize all the experiences to store in the buffer
        for i in range(self._parallel_amount):
            if self._episode_done_previous_step[i]:
                continue
            # Store data at the index targeted by the pointer
            self._observations_current[self._pointer] = observation_current[i]
            self._observations_next[self._pointer] = observation_next[i]
            self._actions[self._pointer] = action[i]
            self._rewards[self._pointer] = reward[i]
            self._episode_done_flags[self._pointer] = episode_done[i]
            # Update the pointer (make sure the first inserted elements are removed first when capacity is exceeded)
            self._pointer = (self._pointer + 1) % self._capacity
            # Update the size with respect to capacity
            self._size = min(self._size + 1, self._capacity)
        # Update the stored episode done flags
        self._episode_done_previous_step = episode_done.copy()

    def get(self,
            amount: int = 0) -> []:
        """
        Get a batch of data from the buffer of the given size. If size is not given all the buffer is used. The buffer
        is not cleared of the data being sampled.

        :param amount: the batch size of data to get
        :return a list containing the arrays of: current observations, actions, rewards, next observations and last step flags
        """
        # Adjust the amount with respect to the buffer current size
        if amount <= 0:
            amount = self._size
        # Return a set of random samples of as large as the given amount
        random_indexes: numpy.ndarray = numpy.random.randint(0, self._size, size=amount)
        return [self._observations_current[random_indexes], self._actions[random_indexes], self._rewards[random_indexes], self._observations_next[random_indexes], self._episode_done_flags[random_indexes]]

    def finish_trajectory(self):
        """
        Finish the trajectory, resetting episode done flags.
        """
        #  Reset stored episode done flags
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    @property
    def capacity(self) -> int:
        """
        The capacity of the buffer.

        :return: the integer capacity of the buffer
        """
        return self._capacity

    @property
    def size(self) -> int:
        """
        The size of the buffer at the current time.

        :return: the integer size of the buffer
        """
        return self._size


class Estimator:
    """
    Estimator defining the real Deterministic Policy Gradient (DDPG) model. It is used to define two identical models:
    target network and main network.

    It is generated given the shape of the observation and action spaces and the hidden layer config defining the
    hidden layers of the network. Hidden layers can be defined separately for the actor network (the policy stream) and
    for the critic network (the q-stream).
    """

    def __init__(self,
                 scope: str,
                 observation_space_shape: (), agent_action_space_shape: (),
                 actor_hidden_layers_config: Config, critic_hidden_layers_config: Config):
        self.scope: str = scope
        with tensorflow.variable_scope(self.scope):
            # Define observations placeholder as a float adaptable array with shape Nx(O) where N is the number of examples and (O) the shape of the observation space
            self.observations = tensorflow.placeholder(shape=(None, *observation_space_shape), dtype=tensorflow.float32, name="observations")
            # Define the actions placeholder with adaptable size Nx(A) where N is the number of examples and (A) the shape of the action space
            self.actions = tensorflow.placeholder(shape=(None, *agent_action_space_shape), dtype=tensorflow.float32, name="actions")
            # Define the policy stream
            with tensorflow.variable_scope("policy_stream"):
                # Define the estimator network policy stream hidden layers from the config
                policy_stream_hidden_layers_output = actor_hidden_layers_config.apply_hidden_layers(self.observations)
                # Define actions as an array of neurons with same adaptable size Nx(A) and with linear activation functions, clipped appropriately
                self.actions_predicted = tensorflow.layers.dense(policy_stream_hidden_layers_output, *agent_action_space_shape, activation=tensorflow.nn.tanh, name="actions_predicted")
            # Define the q-stream (both for given actions and predicted actions)
            with tensorflow.variable_scope("q_stream"):
                # Define the estimator network q-stream hidden layers from the config for the target actions
                q_stream_actions_hidden_layers_output = critic_hidden_layers_config.apply_hidden_layers(tensorflow.concat([self.observations, self.actions], axis=-1))
                # Define the estimator network predicted q-values over the target actions
                self.q_values_actions = tensorflow.squeeze(tensorflow.layers.dense(q_stream_actions_hidden_layers_output, 1), axis=1, name="q_values_actions")
                # Define the estimator network q-stream hidden layers from the config for the predicted actions (automatically shared with targets' one)
                q_stream_predictions_hidden_layers_output = critic_hidden_layers_config.apply_hidden_layers(tensorflow.concat([self.observations, self.actions_predicted], axis=-1))
                # Define the estimator network predicted q-values over the predicted actions
                self.q_values_predictions = tensorflow.squeeze(tensorflow.layers.dense(q_stream_predictions_hidden_layers_output, 1), axis=1, name="q_values_predictions")
            # Define the estimator weight parameters
            self.weight_parameters = [variable for variable in tensorflow.trainable_variables() if self.scope in variable.name]
            self.weight_parameters = sorted(self.weight_parameters, key=lambda parameter: parameter.name)


class DeepDeterministicPolicyGradient(Model):
    """
    Deep Deterministic Policy Gradient Model (DDPG) with FIFO experience replay buffer.
    The model is a deep neural network which hidden layers can be defined by a config parameter. It is an actor-critic
    model. Hidden layers can be defined separately for the actor network (the policy stream) and for the critic
    network (the q-stream). It uses a target network and a main network to correctly evaluate the expected future
    reward in order to stabilize learning.

    In order to synchronize the target network and the main network, every some interval steps the weight have to be
    copied from the main network to the target network.

    This vanilla DDPG implementation tends to be very unstable and its performance varies greatly with respect to the
    hyperparameters. Usually the problem lies in the too high q-values estimations (going to +infinity with time) and
    it is a problem of the vanilla DDPG algorithm as a whole. To help stabilize learning, a SGD optimizer is used
    instead of Adan.

    Note: actions are predicted in the range [-1.0, 1.0] (by using tanh). A rescale is done with respect to lower and
    upper bound if a set of possible actions is given. If not given, the output will remain between [-1.0, 1.0].

    Supported observation spaces:
        - discrete
        - continuous

    Supported action spaces:
        - continuous
    """

    def __init__(self,
                 name: str,
                 actor_hidden_layers_config: Config, critic_hidden_layers_config: Config,
                 learning_rate_policy: float = 1e-3, learning_rate_q_values: float = 1e-3,
                 discount_factor: float = 0.99, polyak_value: float = 0.995,
                 buffer_capacity: int = 1000000,
                 error_clipping: bool = False,
                 huber_delta: float = 1.0):
        # Define model attributes
        self.learning_rate_policy: float = learning_rate_policy
        self.learning_rate_q_values: float = learning_rate_q_values
        self.discount_factor: float = discount_factor
        self.polyak_value: float = polyak_value
        # Define internal model attributes
        self._buffer_capacity: int = buffer_capacity
        self._actor_hidden_layers_config: Config = actor_hidden_layers_config
        self._critic_hidden_layers_config: Config = critic_hidden_layers_config
        self._error_clipping: bool = error_clipping
        self._huber_delta: float = huber_delta
        # Define model empty attributes
        self.buffer: Buffer or None = None
        # Define internal model empty attributes
        self._target_network: Estimator or None = None
        self._main_network: Estimator or None = None
        self._main_network_observations = None
        self._main_network_actions = None
        self._main_network_actions_predicted = None
        self._main_network_q_values_predictions = None
        self._main_network_q_values_actions = None
        self._target_network_observations = None
        self._target_network_q_values_predictions = None
        self._target_q_values = None
        self._policy_stream_loss = None
        self._q_stream_absolute_error = None
        self._q_stream_loss = None
        self._policy_stream_optimizer = None
        self._q_stream_optimizer = None
        self._weight_copier = None
        self._weight_updater = None
        # Generate the base model
        super(DeepDeterministicPolicyGradient, self).__init__(name)
        # Define the types of allowed observation and action spaces
        self._supported_observation_space_types.append(SpaceType.discrete)
        self._supported_observation_space_types.append(SpaceType.continuous)
        self._supported_action_space_types.append(SpaceType.continuous)

    def _define_graph(self):
        # Set the buffer
        self.buffer = Buffer(self._buffer_capacity, self._parallel, self._observation_space_shape, self._agent_action_space_shape)
        # Define two estimator, one for target network and one for main network, with identical structure
        full_scope: str = self._scope + "/" + self._name
        self._main_network = Estimator(full_scope + "/MainNetwork",
                                       self._observation_space_shape, self._agent_action_space_shape,
                                       self._actor_hidden_layers_config, self._critic_hidden_layers_config)
        self._target_network = Estimator(full_scope + "/TargetNetwork",
                                         self._observation_space_shape, self._agent_action_space_shape,
                                         self._actor_hidden_layers_config, self._critic_hidden_layers_config)
        # Assign main and target networks to the model attributes
        self._main_network_observations = self._main_network.observations
        self._main_network_actions = self._main_network.actions
        self._main_network_actions_predicted = self._main_network.actions_predicted
        self._main_network_q_values_predictions = self._main_network.q_values_predictions
        self._main_network_q_values_actions = self._main_network.q_values_actions
        self._target_network_observations = self._target_network.observations
        self._target_network_q_values_predictions = self._target_network.q_values_predictions
        # Define the shared part of the model
        with tensorflow.variable_scope(full_scope):
            # Define target q-values placeholders as adaptable arrays with shape N where N is the number of examples
            self._target_q_values = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="target_q_values")
            # Define policy stream loss
            self._policy_stream_loss = -tensorflow.reduce_mean(self._main_network_q_values_predictions, name="policy_stream_loss")
            # Define the absolute error over the q-stream
            self._q_stream_absolute_error = tensorflow.abs(self._main_network_q_values_actions - self._target_q_values, name="q_stream_absolute_error")
            # Define the q-stream loss with error clipping (huber loss) if required, otherwise mean square error loss
            if self._error_clipping:
                # Compute the Huber loss
                self._q_stream_loss = tensorflow.reduce_mean(tensorflow.where(self._q_stream_absolute_error < self._huber_delta,
                                                                              0.5 * tensorflow.square(self._q_stream_absolute_error),
                                                                              self._q_stream_absolute_error - 0.5), name="q_stream_loss")
            else:
                # Compute MSE loss over the q-stream
                self._q_stream_loss = tensorflow.reduce_mean(tensorflow.square(self._q_stream_absolute_error), name="q_stream_loss")
            # Define the optimizer
            self._policy_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_policy).minimize(self._policy_stream_loss, var_list=[x for x in tensorflow.global_variables() if "MainNetwork/policy_stream" in x.name])
            self._q_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_q_values).minimize(self._q_stream_loss, var_list=[x for x in tensorflow.global_variables() if "MainNetwork/q_stream" in x.name])
            # Define the initializer
            self._initializer = tensorflow.variables_initializer(tensorflow.get_collection(tensorflow.GraphKeys.GLOBAL_VARIABLES, full_scope), name="initializer")
            # Define target network copy/update operations using main network weights
            self._weight_copier = []
            self._weight_updater = []
            for main_network_parameter, target_network_parameter in zip(self._main_network.weight_parameters, self._target_network.weight_parameters):
                # Define the weight copy operation (to copy weights from main network to target network at the start)
                copy_operation = tensorflow.assign(target_network_parameter, main_network_parameter)
                self._weight_copier.append(copy_operation)
                # Define the weight update operation (to update weight from main network to target network during training using Polyak averaging)
                update_operation = tensorflow.assign(target_network_parameter, self.polyak_value * target_network_parameter + (1 - self.polyak_value) * main_network_parameter)
                self._weight_updater.append(update_operation)

    def get_predicted_action_values(self,
                                    session,
                                    observation_current: numpy.ndarray):
        """
        Get the predicted actions values (q-values) according to the model q-stream at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :return: the predicted action values (q-values) predicted by the model
        """
        # Generate a one-hot encoded version of the observation if observation space is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observation_current).reshape(-1)]
        # Return the predicted q-values given the current observation
        return session.run(self._main_network_q_values_predictions,
                           feed_dict={
                               self._main_network_observations: observation_current
                           })

    def get_best_action(self,
                        session,
                        observation_current: numpy.ndarray,
                        possible_actions: [] = None):
        """
        Get the action predicted by the model at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :param possible_actions: the optional list used to bound the actions predicted
        :return: the action predicted by the model
        """
        # Generate a one-hot encoded version of the observation if observation space is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observation_current).reshape(-1)]
        # Get the predicted action given the current observation
        action = session.run(self._main_network_actions_predicted,
                             feed_dict={
                                   self._main_network_observations: observation_current
                             })
        # If there is a possible action array set the boundaries and scale the action
        if possible_actions is not None:
            lower_bound: numpy.ndarray = possible_actions[:, 0]
            upper_bound: numpy.ndarray = possible_actions[:, 1]
            # Rescale the predicted action in the given boundaries
            for i in range(self._parallel):
                action[i] = (upper_bound[i] - lower_bound[i]) * ((action[i] - (-1)) / (1 - (-1))) + lower_bound[i]
        # Return the rescaled (if necessary) action
        return action

    def get_best_action_and_predicted_action_values(self,
                                                    session,
                                                    observation_current: numpy.ndarray,
                                                    possible_actions: [] = None):
        """
        Get the best action predicted by the model at the given current observation and the predicted action values
        (q-values) according to the model at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :param possible_actions: the optional list used to bound the actions predicted
        :return: the best action predicted by the model and the action values (q-values) predicted by the model
        """
        # Return the best action and all predicted actions values
        return self.get_best_action(session, observation_current, possible_actions), self.get_predicted_action_values(session, observation_current)

    def weights_copy(self,
                     session):
        """
        Copy the weights from the main network to the target network.

        :param session: the session of tensorflow currently running
        """
        # Run all the weight copy operations
        session.run(self._weight_copier)

    def weights_update(self,
                       session):
        """
        Update the weights from the main network to the target network by Polyak averaging.

        :param session: the session of tensorflow currently running
        """
        # Run the weight updater
        session.run(self._weight_updater)

    def update(self,
               session,
               batch: []):
        """
        Update the model weights (thus training the model) of the policy and q-stream, given a batch of samples.

        :param session: the session of tensorflow currently running
        :param batch: a batch of samples each one consisting in a tuple of observation current, action, reward, observation next and episode done flags
        :return: the q-stream loss and the policy stream loss
        """
        # Unpack the batch into numpy arrays
        observations_current, actions, rewards, observations_next, episode_done_flags = batch[0], batch[1], batch[2], batch[3], batch[4]
        # Generate a one-hot encoded version of the observations if space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observations_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observations_current.reshape(-1)]
            observations_next: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observations_next.reshape(-1)]
        # Get the predicted target q-values over the current observations
        q_values_next = session.run([self._target_network_q_values_predictions],
                                    feed_dict={
                                        self._target_network_observations: observations_next
                                    })[0]
        # Generate the real target values using the Q-Learning Bellman equation
        q_values_target = rewards + self.discount_factor * (1 - episode_done_flags) * q_values_next
        # Q-stream update
        _, q_stream_loss = session.run([self._q_stream_optimizer, self._q_stream_loss],
                                       feed_dict={
                                            self._main_network_observations: observations_current,
                                            self._target_q_values: q_values_target,
                                            self._main_network_actions: actions,
                                       })
        # Policy stream update
        _, policy_stream_loss = session.run([self._policy_stream_optimizer, self._policy_stream_loss],
                                            feed_dict={
                                                    self._main_network_observations: observations_current,
                                            })
        # Return both losses
        return q_stream_loss, policy_stream_loss

    @property
    def warmup_steps(self) -> int:
        return self._buffer_capacity



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import scripts

from .config import Config, LayerType
from .environment import Environment, SpaceType
from .interface import Interface
from .agent import Agent
from .model import Model
from .volley import Volley
from .experiment import Experiment




#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import tensorflow

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.models import DoubleDeepQLearning
from usienarl.utils import softmax


class DoubleDeepQLearningAgentDirichlet(Agent):
    """
    Double Deep Q-Learning agent with Dirichlet sampling exploration policy.
    It is supplied with a Double Deep Q-Learning model (DDQN).

    The weight copy step interval defines after how many steps per interval the target network weights should be
    updated with the main network weights.

    The batch size define how many steps from the prioritized experience replay built-in buffer should be fed into
    the model when updating. The agent uses a 1-step temporal difference algorithm, i.e. it is updated at every step.

    During training, agent samples from the probability distribution of its computed values and the dirichlet probability
    distribution with given alpha, according to a trade-off parameter which is updated at the end of each episode.

    Note: default model is suited for simple environment without complex dynamics (for example, without an action mask).
    It is advised to implement a custom model for research task using default ones as templates.
    """

    def __init__(self,
                 name: str,
                 model: DoubleDeepQLearning,
                 summary_save_every_steps: int = 500,
                 weight_copy_every_steps: int = 100,
                 batch_size: int = 64,
                 alpha: float = 1.0,
                 dirichlet_trade_off_min: float = 0.5, dirichlet_trade_off_max: float = 1.0,
                 dirichlet_trade_off_update: float = 0.001):
        # Define internal attributes
        self._model: DoubleDeepQLearning = model
        self._alpha: float = alpha
        self._dirichlet_trade_off_min: float = dirichlet_trade_off_min
        self._dirichlet_trade_off_max: float = dirichlet_trade_off_max
        self._dirichlet_trade_off_update: float = dirichlet_trade_off_update
        self._summary_save_every_steps: int = summary_save_every_steps
        self._weight_copy_every_steps: int = weight_copy_every_steps
        self._batch_size: int = batch_size
        # Define empty attributes
        self._last_training_step: int = 0
        self._last_summary_save_step: int = 0
        self._last_weight_copy_step: int = 0
        self._dirichlet_trade_off: float or None = None
        # Generate base agent
        super(DoubleDeepQLearningAgentDirichlet, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        # Generate the model and return a flag stating if generation was successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    self._parallel,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset internal attributes
        self._last_training_step = 0
        self._last_summary_save_step = 0
        self._last_weight_copy_step = 0
        # Initialize the model
        self._model.initialize(logger, session)
        # Reset trade-off to its starting value (the min)
        self._dirichlet_trade_off = self._dirichlet_trade_off_min
        # Run the weight copy operation to uniform main and target networks
        self._model.copy_weights(session)

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   inference_step: int, inference_episode: int):
        # Act randomly
        return interface.sample_agent_action(logger, session)

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  inference_step: int, inference_episode: int):
        # Get the possible actions at the current step of the environment by its interface
        possible_actions: [] = interface.possible_agent_actions(logger, session)
        # Act according to dirichlet approach: first get the softmax over all the q-values predicted by the model
        prior_probabilities = softmax(self._model.get_q_values(session, agent_observation_current, possible_actions))
        # Then generate a dirichlet distribution (d) with parameter alpha
        dirichlet_probabilities: numpy.ndarray = numpy.empty(prior_probabilities.shape)
        for i in range(prior_probabilities.shape[0]):
            dirichlet_probabilities[i] = numpy.random.dirichlet(self._alpha * numpy.ones(prior_probabilities[i].size), 1)
        # Get a random action value (random output) using x * p + (1 - x) * d as probability distribution where x is the trade-off
        # Note: use a multiplicative to adjust dirichlet probabilities accordingly
        multiplicative_mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
        for i in range(self._parallel):
            multiplicative_mask[i, possible_actions[i]] = 1.0
        output = self._dirichlet_trade_off * prior_probabilities + (1 - self._dirichlet_trade_off) * (dirichlet_probabilities * multiplicative_mask)
        # Get an action for each row of the output
        action: numpy.ndarray = numpy.zeros(self._parallel, dtype=int)
        for i in range(self._parallel):
            # Make sure probability rows sums up to 1.0
            probability_row: numpy.ndarray = output[i] / output[i].sum()
            q_value = numpy.random.choice(probability_row, p=probability_row)
            # Set the chosen action as the index of such chosen action value
            action[i] = numpy.argmax(probability_row == q_value)
        # Return the action
        return action

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int):
        # Act with the best policy according to the model
        return self._model.get_action_with_highest_q_value(session, agent_observation_current, interface.possible_agent_actions(logger, session))

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        # Adjust the next observation if the episode is done
        if self._observation_space_type == SpaceType.discrete:
            agent_observation_next[episode_done] = 0
        else:
            agent_observation_next[episode_done] = numpy.zeros(self._observation_space_shape, dtype=float)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), agent_observation_next.copy(), episode_done.copy())

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        # Adjust the next observation if the episode is done
        if self._observation_space_type == SpaceType.discrete:
            agent_observation_next[episode_done] = 0
        else:
            agent_observation_next[episode_done] = numpy.zeros(self._observation_space_shape, dtype=float)
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(),
                                 agent_observation_next.copy(), episode_done.copy())
        # Get the number of steps actually done in the environment according to parallelization
        for step in range(self._last_training_step, train_step):
            # After each weight step interval update the target network weights with the main network weights
            if step - self._last_weight_copy_step >= self._weight_copy_every_steps:
                self._model.copy_weights(session)
                # Update last weight copy step
                self._last_weight_copy_step = step
            # Update the model and get current loss and absolute errors
            loss, absolute_errors = self._model.update(session, self._model.buffer.get(self._batch_size))
            # Update the buffer with absolute error
            self._model.buffer.update(absolute_errors)
            # Save the summary at the current step if it is the appropriate time and if required
            if self._summary_writer is not None:
                if step - self._last_summary_save_step >= self._summary_save_every_steps:
                    # Generate the tensorflow summary on the loss and add it to the writer
                    summary = tensorflow.Summary()
                    summary.value.add(tag="loss", simple_value=loss)
                    self._summary_writer.add_summary(summary, step)
                    # Update last summary save step
                    self._last_summary_save_step = step
        # Update last training step
        self._last_training_step = train_step

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()
        # Increase trade-off by its update value
        self._dirichlet_trade_off = min(self._dirichlet_trade_off_max, self._dirichlet_trade_off + self._dirichlet_trade_off_update)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_episode_current: int,
                                   inference_episode_volley: int):
        pass

    @property
    def saved_variables(self):
        # Return the trainable variables of the agent model in experiment/agent scope
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        # Return the amount of warmup episodes required by the model
        return self._model.warmup_steps



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging

# Import required src

from usienarl import Environment, Agent, Interface


class Volley:
    """
    Base abstract class for a volley. A volley holds data concerning the episodes/steps executed by a certain
    agent in a certain environment.
    """
    def __init__(self,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface,
                 parallel: int):
        # Make sure parameters are correct
        assert (environment is not None and agent is not None and interface is not None and parallel > 0)
        # Define internal attributes
        self._environment: Environment = environment
        self._agent: Agent = agent
        self._interface: Interface = interface
        self._parallel: int = parallel
        # Define empty attributes
        self._number: int or None = None
        self._steps: int or None = None
        self._episodes: int or None = None
        self._start_steps: int or None = None
        self._start_episodes: int or None = None

    def setup(self,
              number: int = 0,
              start_steps: int = 0, start_episodes: int = 0) -> bool:
        """
        Setup the volley, preparing all required values to run. A volley cannot be run without setup first. This also
        initialize the volley.

        :param number: the current number of the volley (usually more volleys are run during an experiment), must be greater or equal than zero
        :param start_steps: the number of steps already executed (in previous volleys), must be greater or equal than zero
        :param start_episodes: the number of episodes already executed (in previous volleys), must be greater or equal than zero
        :return: True if setup is successful, false otherwise
        """
        # Make sure parameters are correct
        assert(number >= 0 and start_steps >= 0 and start_episodes >= 0)
        # Set current steps and episodes counters to their starting value
        self._number = number
        self._start_steps = start_steps
        self._start_episodes = start_episodes
        self._steps = 0
        self._episodes = 0
        # Initialize and return if it is successful
        return self._initialize()

    def _initialize(self) -> bool:
        """
        Initialize the volley attributes.

        :return: True if initialization is successful, false otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def run(self,
            logger: logging.Logger,
            session,
            render: bool = False):
        """
        Run the volley.

        :param logger: the logger used to print the volley information, warnings and errors
        :param session: the session of tensorflow currently running
        :param render: flag defining whether to render the environment or not at every step
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def parallel(self) -> int:
        """
        Number of parallel episodes run by the volley.
        """
        return self._parallel

    @property
    def number(self) -> int or None:
        """
        Number of the volley, i.e. 1st volley (number 0), 2nd volley (number 1), etc.
        It is None if volley is not setup.
        """
        return self._number

    @property
    def steps(self) -> int or None:
        """
        The number of steps run after current setup of the volley (considering as step zero the first one run after
        setup).
        It is None if volley is not setup.
        """
        return self._steps

    @property
    def episodes(self) -> int or None:
        """
        The number of episodes run after current setup of the volley (considering as step zero the first one run after
        setup).
        It is None if volley is not setup.
        """
        return self._episodes

    @property
    def start_steps(self) -> int or None:
        """
        The number of steps already executed when current setup of the volley is done.
        It is None if volley is not setup.
        """
        return self._start_steps

    @property
    def start_episodes(self) -> int or None:
        """
        The number of episodes already executed when current setup of the volley is done.
        It is None if volley is not setup.
        """
        return self._start_episodes



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import tensorflow
import logging

# Import required src

from usienarl import SpaceType


class Model:
    """
    Base abstract model class.
    A model is whatever is moving an agent, deciding actions according to one or more policies and being updated to
    better learn those policies.

    Before use, a model needs to be generated. This define the tensorflow graph beneath the model.

    You should always define your own model, the base class cannot be used as-is.
    """

    def __init__(self,
                 name: str):
        # Define attributes
        self._name: str = name
        # Define empty attributes
        self._initializer = None
        self._scope: str or None = None
        self._parallel: int or None = None
        self._observation_space_type: SpaceType or None = None
        self._observation_space_shape: () = None
        self._agent_action_space_type: SpaceType or None = None
        self._agent_action_space_shape: () = None
        self._supported_observation_space_types: [] = []
        self._supported_action_space_types: [] = []

    def generate(self,
                 logger: logging.Logger,
                 scope: str,
                 parallel: int,
                 observation_space_type: SpaceType, observation_space_shape: (),
                 agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        """
        Generate the tensorflow model with the given scope.

        :param logger: the logger used to print the model information, warnings and errors
        :param scope: the scope of the graph to prepend to the model name
        :param parallel: the amount of parallel episodes run by the model
        :param observation_space_type: the type of the agent observation space
        :param observation_space_shape: the shape of the agent observation space wrapped in a tuple
        :param agent_action_space_type: the type of the agent action space
        :param agent_action_space_shape: the shape of the agent action space wrapped in a tuple
        :return True if the model generation is successful, False otherwise
        """
        # Make sure parameters are valid
        assert(scope is not None and scope)
        assert(parallel > 0)
        assert(observation_space_shape is not None and len(observation_space_shape) > 0)
        assert(agent_action_space_shape is not None and len(observation_space_shape) > 0)
        # Print info
        logger.info("Generating model " + self._name + " with scope " + scope + "...")
        # Set model attributes
        self._observation_space_type = observation_space_type
        self._observation_space_shape = observation_space_shape
        self._agent_action_space_type = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        self._scope = scope
        self._parallel = parallel
        # Check whether or not the observation space and the action space types are supported by the model
        observation_space_type_supported: bool = False
        for space_type in self._supported_observation_space_types:
            if self._observation_space_type == space_type:
                observation_space_type_supported = True
                break
        if not observation_space_type_supported:
            logger.error("Error during generation of model: observation space type not supported")
            return False
        action_space_type_supported: bool = False
        for space_type in self._supported_action_space_types:
            if self._agent_action_space_type == space_type:
                action_space_type_supported = True
                break
        if not action_space_type_supported:
            logger.error("Error during generation of model: action space type not supported")
            return False
        logger.info("Model generation successful")
        # Define the tensorflow model graph
        self._define_graph()
        # Returns True to state success
        return True

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the tensorflow model.

        :param logger: the logger used to print the model information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Initialize the model running the session on the appropriate tensorflow operation
        session.run(self._initializer)
        logger.info("Model initialized to default state")

    def _define_graph(self):
        """
        Define the tensorflow graph of the model.
        It uses as scope the format: experiment_name/agent_name/model_name
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """
        The name of the model.
        """
        return self._name

    @property
    def scope(self) -> str or None:
        """
        The scope of the model.
        It is None if model is not generated.
        """
        return self._scope

    @property
    def parallel(self) -> int or None:
        """
        The number of parallel episodes run by the model.
        It is None if model is not generated.
        """
        return self._parallel

    @property
    def observation_space_type(self) -> SpaceType or None:
        """
        The type of the observation space of the model. It is the same of the encompassing agent.
        It is None if model is not generated.
        """
        return self._observation_space_type

    @property
    def observation_space_shape(self) -> () or None:
        """
        The shape of the observation space of the model. It is the same of the encompassing agent.
        Note: it may differ from the environment's state space shape.
        It is None if model is not generated.
        """
        return self._observation_space_shape

    @property
    def action_space_type(self) -> SpaceType or None:
        """
        The type of the action space of the model. It is the same of the encompassing agent.
        It is None if model is not generated.
        """
        return self._agent_action_space_type

    @property
    def action_space_shape(self) -> () or None:
        """
        The shape of the action space of the model. It is the same of the encompassing agent.
        Note: it may differ from the environment's action space shape.
        It is None if model is not generated.
        """
        return self._agent_action_space_shape

    @property
    def trainable_variables(self):
        """
        The trainable variables in the mode, under 'scope/name'.
        """
        # Get the training variables of the model under its scope: usually, the training variables of the tensorflow graph
        return tensorflow.trainable_variables(self._scope + "/" + self._name + "/")

    @property
    def warmup_steps(self) -> int:
        """
        The number of steps required to warm-up the model.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import tensorflow
import numpy

# Import required src

from usienarl import Interface, SpaceType


class Agent:
    """
    Base agent abstract class.

    An agent defines who or what operates in a certain environment during a certain experiment.
    An agent can act and being updated, but the inner working (model, policies and so on) are left to be decided by the
    implementation.

    An agent needs to be generated before running, and generation is done when executing agent setup.

    Any agent can act in three different modes:
        - warmup mode, before training (not all agents required that, it depends on the agent inner model)
        - train mode, during training
        - inference mode, when exploiting

    One or more policies can be defined for each one of these modes.

    To define your own agent, implement the abstract class in a specific child class.
    """

    def __init__(self,
                 name: str):
        # Define internal attributes
        self._name: str = name
        # Define empty attributes
        self._saver = None
        self._scope: str or None = None
        self._summary_writer = None
        self._parallel: int or None = None
        self._summary_path: str or None = None
        self._save_path: str or None = None
        self._saves_to_keep: int or None = None
        self._observation_space_type: SpaceType or None = None
        self._observation_space_shape = None
        self._agent_action_space_type: SpaceType or None = None
        self._agent_action_space_shape = None
        self._save_counter: int or None = None

    def setup(self,
              logger: logging.Logger,
              scope: str,
              parallel: int,
              observation_space_type: SpaceType, observation_space_shape: (),
              agent_action_space_type: SpaceType, agent_action_space_shape: (),
              summary_path: str = None, save_path: str = None, saves_to_keep: int = 0) -> bool:
        """
        Setup the agent, preparing all its components for execution.

        :param logger: the logger used to print the agent information, warnings and errors
        :param scope: the experiment scope encompassing the agent scope
        :param parallel: the amount of parallel episodes run by the experiment, must be greater than zero
        :param observation_space_type: the space type of the observation space
        :param observation_space_shape: the shape of the observation space, as a tuple
        :param agent_action_space_type: the space type of the agent action space
        :param agent_action_space_shape: the shape of the agent action space, as a tuple
        :param summary_path: the optional path of the summary writer of the agent
        :param save_path: the optional path where to save metagraphs checkpoints of the agent's model
        :param saves_to_keep: the optional number of checkpoint saves to keep, it does nothing if there is no save path
        :return: True if setup is successful, False otherwise
        """
        # Make sure parameters are correct
        assert(parallel > 0 and saves_to_keep >= 0)
        logger.info("Setup of agent " + self._name + " with scope " + scope + "...")
        # Reset agent attributes
        self._scope = scope
        self._parallel = parallel
        self._summary_path = summary_path
        self._save_path = save_path
        self._saves_to_keep = saves_to_keep
        self._observation_space_type: SpaceType = observation_space_type
        self._observation_space_shape = observation_space_shape
        self._agent_action_space_type: SpaceType = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        # Try to generate the agent inner model
        if not self._generate(logger,
                              observation_space_type, observation_space_shape,
                              agent_action_space_type, agent_action_space_shape):
            return False
        # Define the summary writer if required
        if summary_path is not None:
            self._summary_writer = tensorflow.summary.FileWriter(summary_path, graph=tensorflow.get_default_graph())
            logger.info("A Tensorboard summary for the agent will be updated during training of its internal model")
            logger.info("Tensorboard summary path: " + summary_path)
        # Define the saver if required
        if self._save_path is not None and self._save_path and self._saves_to_keep > 0:
            self._saver = tensorflow.train.Saver(self.saved_variables, max_to_keep=self._saves_to_keep)
            if self._saves_to_keep > 1:
                logger.info("Agent model metagraph will be saved after each training/validation pair. A set of " + str(self._saves_to_keep) + " models will be stored.")
            else:
                logger.info("Agent model metagraph will be saved after each training/validation pair")
            logger.info("Agent model metagraphs are saved at " + self._save_path)
            self._save_counter: int = 0
        # Validate setup
        return True

    def restore(self,
                logger: logging.Logger,
                session,
                path: str) -> bool:
        """
        Restore the agent's model from the checkpoint at the given path.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param path: the path from which to restore, it is required
        :return: True if restore is successful, false otherwise
        """
        # Make sure parameters are correct
        assert(path is not None and path)
        # Get checkpoint from path
        checkpoint = tensorflow.train.get_checkpoint_state(path)
        # If checkpoint exists restore from checkpoint
        if checkpoint and checkpoint.model_checkpoint_path:
            self._saver.restore(session, tensorflow.train.latest_checkpoint(path))
            logger.info("Model graph stored at " + path + " loaded successfully!")
            return True
        logger.error("Checkpoint path specified is wrong: no model can be accessed at " + path)
        return False

    def save(self,
             logger: logging.Logger,
             session):
        """
        Save the agent's model metagraph. It does nothing if a saver is not defined.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Check if the saver exists
        if self._saver is None:
            return
        logger.info("Saving the agent " + self._name + " metagraph at path " + self._save_path + "...")
        self._saver.save(session, self._save_path, self._save_counter)
        self._save_counter += 1
        logger.info("Agent " + self._name + " metagraph saved successfully")

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        """
        Generate the agent's model. Used to generate all custom components of the agent.
        It is always called during setup.

        :param logger: the logger used to print the agent information, warnings and errors
        :param observation_space_type: the space type of the observation space
        :param observation_space_shape: the shape of the observation space, as a tuple
        :param agent_action_space_type: the space type of the agent action space
        :param agent_action_space_shape: the shape of the agent action space, as a tuple
        :return: True if setup is successful, False otherwise
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the agent before acting in the environment.
        The environment at this stage is already initialized.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   warmup_step: int, warmup_episode: int) -> numpy.ndarray:
        """
        Take an action given the current agent observation in warmup mode.
        Usually it uses a random policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent wrapped in a numpy array
        :param warmup_step: the current absolute warm-up step of the experiment the agent is warming-up into
        :param warmup_episode: the current absolute warm-up episode of the experiment the agent is warming-up into
        :return: the action decided by the agent wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  train_step: int, train_episode: int) -> numpy.ndarray:
        """
        Take an action given the current agent observation in train mode.
        Usually it uses an exploring policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent wrapped in a numpy array
        :param train_step: the current absolute train step of the experiment the agent is training into
        :param train_episode: the current absolute train episode of the experiment the agent is training into
        :return: the action decided by the agent wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int) -> numpy.ndarray:
        """
        Take an action given the current agent observation in validation/test mode.
        Usually it uses the best possible policy.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent wrapped in a numpy array
        :param inference_step: the current absolute validation/test step of the experiment the agent is executing into
        :param inference_episode: the current absolute validation/test episode of the experiment the agent is executing into
        :return: the action decided by the agent wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        """
        Complete a warm-up step.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent wrapped in a numpy array
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation wrapped in a numpy array
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between wrapped in a numpy array
        :param episode_done: the episode done flag raised by the environment wrapped in a numpy array
        :param agent_observation_next: the next observation of the agent wrapped in a numpy array
        :param warmup_step: the current absolute warm-up step of the experiment the agent is warming-up into
        :param warmup_episode: the current absolute warm-up episode of the experiment the agent is warming-up into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        """
        Complete a train step.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent wrapped in a numpy array
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation wrapped in a numpy array
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between wrapped in a numpy array
        :param episode_done: the episode done flag raised by the environment wrapped in a numpy array
        :param agent_observation_next: the next observation of the agent wrapped in a numpy array
        :param train_step: the current absolute train step of the experiment the agent is training into
        :param train_episode: the current absolute train episode of the experiment the agent is training into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        """
        Complete an validation/test step..

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param agent_observation_current: the current observation of the agent
        :param agent_action: the action taken in the environment as seen by the agent leading from the current observation to the next observation
        :param reward: the reward obtained by the combination of current observation, next observation and action in-between
        :param episode_done: the episode done flag raised by the environment wrapped in a numpy array
        :param agent_observation_next: the next observation of the agent
        :param inference_step: the current absolute validation/test step of the experiment the agent is executing into
        :param inference_episode: the current absolute validation/test episode of the experiment the agent is executing into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        """
        Finish a warm-up episode.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment  wrapped in a numpy array
        :param last_step_reward: the reward obtained in the last step in the passed episode wrapped in a numpy array  wrapped in a numpy array
        :param episode_total_reward: the reward obtained in the passed episode wrapped in a numpy array
        :param warmup_step: the current absolute warm-up step of the experiment the agent is warming-up into
        :param warmup_episode: the current absolute warm-up episode of the experiment the agent is warming-up into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        """
        Finish a train episode.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param last_step_reward: the reward obtained in the last step in the passed episode  wrapped in a numpy array
        :param episode_total_reward: the reward obtained in the passed episode  wrapped in a numpy array
        :param train_step: the current absolute train step of the experiment the agent is training into
        :param train_episode: the current absolute train episode of the experiment the agent is training into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_step: int, inference_episode: int):
        """
        Finish an validation/test episode.

        :param logger: the logger used to print the agent information, warnings and errors
        :param session: the session of tensorflow currently running
        :param interface: the interface between the agent and the environment
        :param last_step_reward: the reward obtained in the last step in the passed episode wrapped in a numpy array
        :param episode_total_reward: the reward obtained in the passed episode wrapped in a numpy array
        :param inference_step: the current absolute validation/test step of the experiment the agent is executing into
        :param inference_episode: the current absolute validation/test episode of the experiment the agent is executing into
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """
        The name of the agent.
        """
        return self._name

    @property
    def scope(self) -> str or None:
        """
        The scope of the agent.
        It is None if agent is not setup.
        """
        return self._scope

    @property
    def parallel(self) -> int or None:
        """
        The number of parallel episodes run by the agent.
        It is None if agent is not setup.
        """
        return self._parallel

    @property
    def observation_space_type(self) -> SpaceType or None:
        """
        The type of the observation space of the agent.
        It is None if agent is not setup.
        """
        return self._observation_space_type

    @property
    def observation_space_shape(self) -> () or None:
        """
        The shape of the observation space of the agent.
        Note: it may differ from the environment's state space shape.
        It is None if agent is not setup.
        """
        return self._observation_space_shape

    @property
    def action_space_type(self) -> SpaceType or None:
        """
        The type of the action space of the agent.
        It is None if agent is not setup.
        """
        return self._agent_action_space_type

    @property
    def action_space_shape(self) -> () or None:
        """
        The shape of the action space of the agent.
        Note: it may differ from the environment's action space shape.
        It is None if agent is not setup.
        """
        return self._agent_action_space_shape

    @property
    def summary_path(self) -> str or None:
        """
        The path of the Tensorboard summaries saved by the agent.
        It is None if agent is not setup.
        """
        return self._summary_path

    @property
    def save_path(self) -> str or None:
        """
        The path of the checkpoint metagraph saves of the agent.
        It is None if agent is not setup.
        """
        return self._save_path

    @property
    def saved_variables(self):
        """
        The trainable variables of the agent (usually of its model).
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def warmup_steps(self) -> int:
        """
        The integer number of warm-up steps required by the agent.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy

# Import required src

from usienarl import Interface, Environment, SpaceType


class PassThroughInterface(Interface):
    """
    Basic pass-through interface just connecting an environment to an agent. It does not modify nor the shape nor
    the type of the actions and of the states.

    It is used by default in any experiment where a specific interface is not supplied.
    """

    def __init__(self,
                 environment: Environment):
        # Generate the base interface
        super(PassThroughInterface, self).__init__(environment)

    def agent_action_to_environment_action(self,
                                           logger: logging.Logger,
                                           session,
                                           agent_action: numpy.ndarray) -> numpy.ndarray:
        # Just return the agent action
        return agent_action

    def environment_action_to_agent_action(self,
                                           logger: logging.Logger,
                                           session,
                                           environment_action: numpy.ndarray) -> numpy.ndarray:
        # Just return the environment action
        return environment_action

    def environment_state_to_observation(self,
                                         logger: logging.Logger,
                                         session,
                                         environment_state: numpy.ndarray) -> numpy.ndarray:
        # Just return the environment state
        return environment_state

    def possible_agent_actions(self,
                               logger: logging.Logger,
                               session) -> []:
        # Just return the environment possible actions
        return self._environment.possible_actions(logger, session)

    @property
    def observation_space_type(self) -> SpaceType:
        # Just return the environment state space type
        return self._environment.state_space_type

    @property
    def observation_space_shape(self):
        # Just return the environment state space shape
        return self._environment.state_space_shape

    @property
    def agent_action_space_type(self) -> SpaceType:
        # Just return the environment action space type
        return self._environment.action_space_type

    @property
    def agent_action_space_shape(self):
        # Just return the environment action space shape
        return self._environment.action_space_shape
